from typing import Union, Dict
import pathlib
import logging
import time

from .configuration import _load_config
from .request import ServiceXRequest
from .frontend import ServiceXFrontend
from .output import _output_handler

log = logging.getLogger(__name__)

class DataBinder:
    """ Manage and categorize numerous ServiceX data from a configuration file"""

    def __init__(self, config: Union[str, pathlib.Path]):
        self._config = _load_config(config)
        self._requests = ServiceXRequest(self._config).get_requests()
        self._sx = ServiceXFrontend(self._config, self._requests)
        self._cache_before_requests = self._sx.get_current_cache()
        

    def deliver(self, timer=False) -> Dict:
        
        """Get a list of parquet files for each ServiceX request"""
        t_0 = time.perf_counter()
        try:
            output_parquet_list = self._sx.get_servicex_data()
        except:
            log.exception("Exception occured while gettting data via ServiceX")
            raise
        t_1 = time.perf_counter()

        """Handles ServiceX delivered output"""
        out = _output_handler(self._config, self._requests, output_parquet_list, self._cache_before_requests)
        t_2 = time.perf_counter()

        if timer:
            print(f"---------------- Timer ----------------")
            print(f"ServiceX data delivery: {t_1-t_0:0.1f} seconds")
            print(f"Post-processing       : {t_2-t_1:0.1f} seconds")

        return out


import asyncio
from servicex import ServiceXDataset
from aiohttp import ClientSession
import nest_asyncio
from typing import Any, Dict, List, Optional, Union
from pathlib import Path
import logging

from .output import get_tree_name

log = logging.getLogger(__name__)

class ServiceXFrontend:

    def __init__(self, config: Dict[str, Any], servicex_requests):
        """
        self._list_sx_dataset_query_pair   List of ServiceX dataset and query pair
        """
        self._config = config
        self._servicex_requests = servicex_requests
        self._backend = self._config.get('General')['ServiceXBackendName'].lower()

    def get_current_cache(self) -> List:        
        cache_path = ServiceXDataset("",backend_name=self._config['General']['ServiceXBackendName'])._cache._path
        log.debug(f"local cache path: {cache_path}")
        query_cache_status = Path.joinpath(cache_path, "query_cache_status")
        return list(query_cache_status.glob('*'))

    def get_servicex_data(self, test_run=False):
        """
        Get data from ServiceX
        """
        log.info(f"retrieving data via {self._config['General']['ServiceXBackendName']} ServiceX..")

        if 'IgnoreServiceXCache' in self._config['General'].keys():
            ignoreCache = self._config['General']['IgnoreServiceXCache']
        else:
            ignoreCache = False

        if 'uproot' in self._backend:
            transformer_image = "sslhep/servicex_func_adl_uproot_transformer:develop"
        elif 'xaod' in self._backend:
            transformer_image = "sslhep/servicex_func_adl_xaod_transformer:develop"
        
        nest_asyncio.apply()
 
        async def bound_get_data(sem, sx_ds, query, title):
            async with sem:
                return await sx_ds.get_data_parquet_async(query, title)

        async def _get_my_data():
            sem = asyncio.Semaphore(50) # Limit maximum concurrent ServiceX requests
            tasks = []
            
            async with ClientSession() as session:
                for request in self._servicex_requests:
                    sx_ds = ServiceXDataset(dataset=request['dataset'], \
                                            backend_name=self._config['General']['ServiceXBackendName'], \
                                            # image=transformer_image, \
                                            session_generator=session, \
                                            ignore_cache=ignoreCache)
                    query = request['query']
                    
                    if 'uproot' in self._backend:
                        title = request['Sample'] + ' - ' + get_tree_name(query)
                    else:
                        title = request['Sample']
                        
                    task = asyncio.ensure_future(bound_get_data(sem, sx_ds, query, title))
                    tasks.append(task)
                return await asyncio.gather(*tasks)

        newloop = asyncio.get_event_loop()
        data = newloop.run_until_complete(_get_my_data())
        log.info(f"complete ServiceX data delivery..")
        return data



import yaml
import logging
import pathlib
from typing import Any, Dict, Union

log = logging.getLogger(__name__)

def _load_config(input_config: Union[str, pathlib.Path, Dict[str, Any]]) -> Dict[str, Any]:
    """Loads, validates, and returns a config file from the provided path.
    Args:
        input_config (Union[str, pathlib.Path, Dict[str, Any]]): path to config file or config as dict 
    Returns:
        Dict[str, Any]: configuration
    """
    
    if isinstance(input_config, Dict):
        _validate_config(input_config)
        return input_config
    else:
        file_path = pathlib.Path(input_config)
        log.info(f"opening config file: {file_path}")
        try:
            config = yaml.safe_load(file_path.read_text())
            _validate_config(config)
            return config
        except:
            raise FileNotFoundError(f"Exception occured while reading config file: {file_path}")

def _validate_config(config: Dict[str, Any]) -> bool:
    """Returns True if the config file is validated, otherwise raises exceptions.
    Checks that the config satisfies the json schema, and performs additional checks to
    validate the config further.
    Args:
        config (Dict[str, Any]): configuration
    Raises:
        NotImplementedError
        ValueError
        KeyError
    Returns:
        bool: whether the validation was successful
    """
    
    if 'General' not in config.keys() and 'Sample' not in config.keys():
        raise KeyError(f"You should have 'General' and 'Sample' in the config")
    
    if 'ServiceXBackendName' not in config['General'].keys():
        raise KeyError(f"ServiceXBackendName is required")
    elif 'uproot' not in config['General']['ServiceXBackendName'].lower() and \
         'xaod' not in config['General']['ServiceXBackendName'].lower():
        raise ValueError(f"ServiceXBackendName should contain either uproot or xaod")

    if 'OutputDirectory' not in config['General'].keys():
        raise KeyError(f"OutputDirectory is required")

    if 'OutputFormat' not in config['General'].keys():
        raise KeyError(f"OutputFormat is required")
    elif config['General']['OutputFormat'].lower() != 'parquet' and \
         config['General']['OutputFormat'].lower() != 'root':
        raise ValueError(f"OutputFormat can be either parquet or root")

    for sample in config['Sample']:        
        if 'RucioDID' not in sample.keys() and 'XRootDFiles' not in sample.keys():
            raise KeyError(f"Sample {sample['Name']} should have RucioDID")
        if 'RucioDID' in sample.keys():
            for did in sample['RucioDID'].split(","):
                if len(did.split(":")) != 2:
                    raise ValueError(f"Sample {sample['Name']} - RucioDID {did} is missing the scope")
        if 'Tree' in sample and 'uproot' not in config['General']['ServiceXBackendName'].lower():
            raise KeyError(f"Tree in Sample {sample['Name']} is only for uproot backend type")
        if 'Columns' in sample and 'FuncADL' in sample:
            raise KeyError(f"Sample {sample['Name']} - Use one type of query per sample: Columns for TCut and FuncADL for func-adl")
        if 'FuncADL' in sample and 'Filter' in sample:
            raise KeyError(f"Sample {sample['Name']} - You cannot use Filter with func-adl query")

    log.debug("config looks okay")
    return True


import setuptools
import codecs
import os.path


with open("README.md") as fh:
    long_description = fh.read()

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

setuptools.setup(name="servicex_databinder",
                 version=get_version("servicex_databinder/__init__.py"),
                 packages=setuptools.find_packages(exclude=['tests']),
                 description="ServiceX data management using a configuration file",
                 long_description=long_description,
                 long_description_content_type='text/markdown',
                 author="KyungEon Choi (UT Austin)",
                 author_email="kyungeonchoi@utexas.edu",
                 url="https://github.com/kyungeonchoi/ServiceXDataBinder",
                 license="BSD 3-clause",
                 python_requires='>=3.6',
                 install_requires=[
                    "servicex>=2.5",
                    "tcut-to-qastle>=0.7",                    
                    "nest-asyncio>=1.5.1",                
                    "tqdm>=4.60.0",
                    "pyarrow>=3.0.0",
                    "backoff>=1.11.1",
                    "func_adl_servicex>=1.1"
                    ],
                 classifiers=[
                    "Development Status :: 3 - Alpha",
                    "Intended Audience :: Developers",
                    "Intended Audience :: Information Technology",
                    "Intended Audience :: Science/Research",
                    "Programming Language :: Python",
                    "Topic :: Scientific/Engineering :: Physics",
                    "Topic :: Software Development",
                    "Topic :: Utilities",
                    ],
                 platforms="Any",)
